var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
(function (factory) {
    if (typeof module === "object" && typeof module.exports === "object") {
        var v = factory(require, exports);
        if (v !== undefined) module.exports = v;
    }
    else if (typeof define === "function" && define.amd) {
        define(["require", "exports", "./colors"], factory);
    }
})(function (require, exports) {
    "use strict";
    Object.defineProperty(exports, "__esModule", { value: true });
    var colors_1 = require("./colors");
    var Printer = /** @class */ (function () {
        function Printer(config) {
            var _this = this;
            if ((!config || !config.height) &&
                typeof process.stdout.rows !== "number") {
                throw new Error("Could not fall back to a default height, because " +
                    "'process.stdout.rows' is no number. You have to " +
                    "specify 'height' when calling the constructor.");
            }
            if ((!config || !config.width) &&
                typeof process.stdout.columns !== "number") {
                throw new Error("Could not fall back to a default width, because " +
                    "'process.stdout.columns' is no number. You have to " +
                    "specify 'width' when calling the constructor.");
            }
            this.height = (config && config.height) || process.stdout.rows;
            this.width = (config && config.width) || process.stdout.columns;
            this.printer = Array.from({ length: this.height }).map(function (_) {
                return Array.from({ length: _this.width }).map(function (__) { return ({
                    backgroundColor: (config && config.backgroundColor) || null,
                    foregroundColor: (config && config.foregroundColor) || null,
                    value: " ",
                }); });
            });
            this.updateBuffer = [];
        }
        Printer.prototype.clear = function (colors) {
            var _this = this;
            var backgroundColor = (colors && colors.backgroundColor) || null;
            var foregroundColor = (colors && colors.foregroundColor) || null;
            Array.from({ length: this.height }).forEach(function (_, i) {
                Array.from({ length: _this.width }).forEach(function (__, j) {
                    _this.setPixel(i, j, {
                        backgroundColor: backgroundColor,
                        foregroundColor: foregroundColor,
                        value: " ",
                    });
                });
            });
        };
        Printer.prototype.getPixel = function (row, column) {
            return this.printer[row][column];
        };
        Printer.prototype.hideCursor = function () {
            process.stdout.write("\u001B[?25l");
        };
        Printer.prototype.print = function () {
            var print = this.printer
                .map(function (row) {
                return row
                    .map(function (cell) {
                    return "" + colors_1.getTerminalColor("foreground", cell.foregroundColor) + colors_1.getTerminalColor("background", cell.backgroundColor) + cell.value;
                })
                    .join("");
            })
                .join("");
            process.stdout.write("\x1b[1;1H");
            process.stdout.write(print);
        };
        Printer.prototype.setPixel = function (row, column, pixel) {
            if (typeof pixel.value === "string" && pixel.value.length !== 1) {
                throw new Error("Can't set values for a single pixel that have a length which is " +
                    "not equal to 1.");
            }
            this.printer[row][column] = __assign(__assign({}, this.printer[row][column]), pixel);
            var elementInUpdateBuffer = this.updateBuffer.findIndex(function (update) { return update.row === row && update.column === column; });
            if (elementInUpdateBuffer < 0) {
                this.updateBuffer.push({
                    column: column,
                    pixel: {
                        backgroundColor: pixel.backgroundColor || null,
                        foregroundColor: pixel.foregroundColor || null,
                        value: pixel.value || this.printer[row][column].value,
                    },
                    row: row,
                });
            }
            else {
                this.updateBuffer.splice(elementInUpdateBuffer, 1, {
                    column: column,
                    pixel: __assign(__assign({}, this.updateBuffer[elementInUpdateBuffer].pixel), pixel),
                    row: row,
                });
            }
        };
        Printer.prototype.showCursor = function () {
            process.stdout.write("\u001B[?25h");
        };
        Printer.prototype.update = function () {
            var _this = this;
            this.sortUpdateBuffer();
            var updateString = this.updateBuffer.reduce(function (currentUpdateString, updateElement, index) {
                var newUpdateString = currentUpdateString;
                if (index === 0) {
                    // this element is the first, so write position and both colors
                    newUpdateString += "\u001B[" + (updateElement.row +
                        1) + ";" + (updateElement.column + 1) + "H";
                    newUpdateString += colors_1.getTerminalColor("background", updateElement.pixel.backgroundColor || null);
                    newUpdateString += colors_1.getTerminalColor("foreground", updateElement.pixel.foregroundColor || null);
                }
                else {
                    // write position (if necessary)
                    var previousElement = _this.updateBuffer[index - 1];
                    var hasPredecessorInSameRow = updateElement.column > 0 &&
                        _this.updateBuffer[index - 1].row === updateElement.row;
                    if (!hasPredecessorInSameRow) {
                        // there is no update element in this row, so position the
                        // cursor at the location of the update element
                        newUpdateString += "\u001B[" + (updateElement.row +
                            1) + ";" + (updateElement.column + 1) + "H";
                    }
                    else {
                        // there is a update element in this row, so the cursor is
                        // already located in this row
                        var hasDirectPredecessor = hasPredecessorInSameRow &&
                            previousElement.column === updateElement.column - 1;
                        if (!hasDirectPredecessor) {
                            // the current element does not directly follow the last
                            // element, so move the cursor as many columns forward as
                            // necessary
                            newUpdateString += "\u001B[" + (updateElement.column -
                                previousElement.column -
                                1) + "C";
                        }
                    }
                    // write background color (if necessary)
                    if (!_this.equalColors(previousElement.pixel.backgroundColor || null, updateElement.pixel.backgroundColor || null)) {
                        newUpdateString += colors_1.getTerminalColor("background", updateElement.pixel.backgroundColor || null);
                    }
                    // write foreground color (if necessary)
                    if (!_this.equalColors(previousElement.pixel.foregroundColor || null, updateElement.pixel.foregroundColor || null)) {
                        newUpdateString += colors_1.getTerminalColor("foreground", updateElement.pixel.foregroundColor || null);
                    }
                }
                // write value
                newUpdateString += updateElement.pixel.value;
                return newUpdateString;
            }, "");
            process.stdout.write(updateString);
            this.updateBuffer = [];
        };
        Printer.prototype.writeCenteredRow = function (row, text, colors) {
            this.writeRow(row, Math.floor((this.width - text.length) / 2), text, colors);
        };
        Printer.prototype.writeCenteredText = function (text, colors) {
            var _this = this;
            var textArray = text.split("\n");
            var firstRow = Math.floor((this.height - textArray.length) / 2);
            textArray.forEach(function (string, index) {
                _this.writeCenteredRow(firstRow + index, string, colors);
            });
        };
        Printer.prototype.writeRow = function (row, column, text, colors) {
            var _this = this;
            var backgroundColor = (colors && colors.backgroundColor) || null;
            var foregroundColor = (colors && colors.foregroundColor) || null;
            text.split("").forEach(function (value, index) {
                if (column + index < _this.width) {
                    _this.setPixel(row, column + index, {
                        backgroundColor: backgroundColor,
                        foregroundColor: foregroundColor,
                        value: value,
                    });
                }
            });
        };
        Printer.prototype.equalColors = function (firstColor, secondColor) {
            if (firstColor === null && secondColor === null) {
                return true;
            }
            if (firstColor === null || secondColor === null) {
                return false;
            }
            return (colors_1.getTerminalColor("foreground", firstColor) ===
                colors_1.getTerminalColor("foreground", secondColor));
        };
        Printer.prototype.sortUpdateBuffer = function () {
            this.updateBuffer = this.updateBuffer.sort(function (firstElement, secondElement) {
                if (firstElement.row < secondElement.row ||
                    (firstElement.row === secondElement.row &&
                        firstElement.column < secondElement.column)) {
                    return -1;
                }
                /* istanbul ignore else */
                if (firstElement.row > secondElement.row ||
                    (firstElement.row === secondElement.row &&
                        firstElement.column > secondElement.column)) {
                    return 1;
                }
                /* istanbul ignore next */
                return 0;
            });
        };
        return Printer;
    }());
    exports.Printer = Printer;
});
//# sourceMappingURL=index.js.map